"""Synchronous IO wrappers around jeepney
"""
import array
from collections import deque
from errno import ECONNRESET
import functools
from itertools import count
import os
from selectors import DefaultSelector, EVENT_READ
import socket
import time
from typing import Optional
from warnings import warn

from jeepney import Parser, Message, MessageType, HeaderFields
from jeepney.auth import Authenticator, BEGIN
from jeepney.bus import get_bus
from jeepney.fds import FileDescriptor, fds_buf_size
from jeepney.wrappers import ProxyBase, unwrap_msg
from jeepney.routing import Router
from jeepney.bus_messages import message_bus
from .common import MessageFilters, FilterHandle, check_replyable

__all__ = [
    'open_dbus_connection',
    'DBusConnection',
    'Proxy',
]


class _Future:
    def __init__(self):
        self._result = None

    def done(self):
        return bool(self._result)

    def set_exception(self, exception):
        self._result = (False, exception)

    def set_result(self, result): 
        self._result = (True, result)

    def result(self):
        success, value = self._result
        if success:
            return value
        raise value


def timeout_to_deadline(timeout):
    if timeout is not None:
        return time.monotonic() + timeout
    return None

def deadline_to_timeout(deadline):
    if deadline is not None:
        return max(deadline - time.monotonic(), 0.)
    return None


class DBusConnectionBase:
    """Connection machinery shared by this module and threading"""
    def __init__(self, sock: socket.socket, enable_fds=False):
        self.sock = sock
        self.enable_fds = enable_fds
        self.parser = Parser()
        self.outgoing_serial = count(start=1)
        self.selector = DefaultSelector()
        self.select_key = self.selector.register(sock, EVENT_READ)
        self.unique_name = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        return False

    def _serialise(self, message: Message, serial) -> (bytes, Optional[array.array]):
        if serial is None:
            serial = next(self.outgoing_serial)
        fds = array.array('i') if self.enable_fds else None
        data = message.serialise(serial=serial, fds=fds)
        return data, fds

    def _send_with_fds(self, data, fds):
        bytes_sent = self.sock.sendmsg(
            [data], [(socket.SOL_SOCKET, socket.SCM_RIGHTS, fds)]
        )
        # If sendmsg succeeds, I think ancillary data has been sent atomically?
        # So now we just need to send any leftover normal data.
        if bytes_sent < len(data):
            self.sock.sendall(data[bytes_sent:])

    def _receive(self, deadline):
        while True:
            msg = self.parser.get_next_message()
            if msg is not None:
                return msg

            b, fds = self._read_some_data(timeout=deadline_to_timeout(deadline))
            self.parser.add_data(b, fds=fds)

    def _read_some_data(self, timeout=None):
        for key, ev in self.selector.select(timeout):
            if key == self.select_key:
                if self.enable_fds:
                    return self._read_with_fds()
                else:
                    return unwrap_read(self.sock.recv(4096)), []

        raise TimeoutError

    def _read_with_fds(self):
        nbytes = self.parser.bytes_desired()
        data, ancdata, flags, _ = self.sock.recvmsg(nbytes, fds_buf_size())
        if flags & getattr(socket, 'MSG_CTRUNC', 0):
            self.close()
            raise RuntimeError("Unable to receive all file descriptors")
        return unwrap_read(data), FileDescriptor.from_ancdata(ancdata)

    def close(self):
        """Close the connection"""
        self.selector.close()
        self.sock.close()


class DBusConnection(DBusConnectionBase):
    def __init__(self, sock: socket.socket, enable_fds=False):
        super().__init__(sock, enable_fds)

        # Message routing machinery
        self._router = Router(_Future)  # Old interface, for backwards compat
        self._filters = MessageFilters()

        # Say Hello, get our unique name
        self.bus_proxy = Proxy(message_bus, self)
        hello_reply = self.bus_proxy.Hello()
        self.unique_name = hello_reply[0]

    @property
    def router(self):
        warn("conn.router is deprecated, see the docs for APIs to use instead.",
             stacklevel=2)
        return self._router

    def send(self, message: Message, serial=None):
        """Serialise and send a :class:`~.Message` object"""
        data, fds = self._serialise(message, serial)
        if fds:
            self._send_with_fds(data, fds)
        else:
            self.sock.sendall(data)

    send_message = send  # Backwards compatibility

    def receive(self, *, timeout=None) -> Message:
        """Return the next available message from the connection

        If the data is ready, this will return immediately, even if timeout<=0.
        Otherwise, it will wait for up to timeout seconds, or indefinitely if
        timeout is None. If no message comes in time, it raises TimeoutError.
        """
        return self._receive(timeout_to_deadline(timeout))

    def recv_messages(self, *, timeout=None):
        """Receive one message and apply filters

        See :meth:`filter`. Returns nothing.
        """
        msg = self.receive(timeout=timeout)
        self._router.incoming(msg)
        for filter in self._filters.matches(msg):
            filter.queue.append(msg)

    def send_and_get_reply(self, message, *, timeout=None, unwrap=None):
        """Send a message, wait for the reply and return it

        Filters are applied to other messages received before the reply -
        see :meth:`add_filter`.
        """
        check_replyable(message)
        deadline = timeout_to_deadline(timeout)

        if unwrap is None:
            unwrap = False
        else:
            warn("Passing unwrap= to .send_and_get_reply() is deprecated and "
                 "will break in a future version of Jeepney.", stacklevel=2)

        serial = next(self.outgoing_serial)
        self.send_message(message, serial=serial)
        while True:
            msg_in = self.receive(timeout=deadline_to_timeout(deadline))
            reply_to = msg_in.header.fields.get(HeaderFields.reply_serial, -1)
            if reply_to == serial:
                if unwrap:
                    return unwrap_msg(msg_in)
                return msg_in

            # Not the reply
            self._router.incoming(msg_in)
            for filter in self._filters.matches(msg_in):
                filter.queue.append(msg_in)

    def filter(self, rule, *, queue: Optional[deque] =None, bufsize=1):
        """Create a filter for incoming messages

        Usage::

            with conn.filter(rule) as matches:
                # matches is a deque containing matched messages
                matching_msg = conn.recv_until_filtered(matches)

        :param jeepney.MatchRule rule: Catch messages matching this rule
        :param collections.deque queue: Matched messages will be added to this
        :param int bufsize: If no deque is passed in, create one with this size
        """
        if queue is None:
            queue = deque(maxlen=bufsize)
        return FilterHandle(self._filters, rule, queue)

    def recv_until_filtered(self, queue, *, timeout=None) -> Message:
        """Process incoming messages until one is filtered into queue

        Pops the message from queue and returns it, or raises TimeoutError if
        the optional timeout expires. Without a timeout, this is equivalent to::

            while len(queue) == 0:
                conn.recv_messages()
            return queue.popleft()

        In the other I/O modules, there is no need for this, because messages
        are placed in queues by a separate task.

        :param collections.deque queue: A deque connected by :meth:`filter`
        :param float timeout: Maximum time to wait in seconds
        """
        deadline = timeout_to_deadline(timeout)
        while len(queue) == 0:
            self.recv_messages(timeout=deadline_to_timeout(deadline))
        return queue.popleft()


class Proxy(ProxyBase):
    """A blocking proxy for calling D-Bus methods

    You can call methods on the proxy object, such as ``bus_proxy.Hello()``
    to make a method call over D-Bus and wait for a reply. It will either
    return a tuple of returned data, or raise :exc:`.DBusErrorResponse`.
    The methods available are defined by the message generator you wrap.

    You can set a time limit on a call by passing ``_timeout=`` in the method
    call, or set a default when creating the proxy. The ``_timeout`` argument
    is not passed to the message generator.
    All timeouts are in seconds, and :exc:`TimeoutErrror` is raised if it
    expires before a reply arrives.

    :param msggen: A message generator object
    :param ~blocking.DBusConnection connection: Connection to send and receive messages
    :param float timeout: Default seconds to wait for a reply, or None for no limit
    """
    def __init__(self, msggen, connection, *, timeout=None):
        super().__init__(msggen)
        self._connection = connection
        self._timeout = timeout

    def __repr__(self):
        extra = '' if (self._timeout is None) else f', timeout={self._timeout}'
        return f"Proxy({self._msggen}, {self._connection}{extra})"

    def _method_call(self, make_msg):
        @functools.wraps(make_msg)
        def inner(*args, **kwargs):
            timeout = kwargs.pop('_timeout', self._timeout)
            msg = make_msg(*args, **kwargs)
            assert msg.header.message_type is MessageType.method_call
            return unwrap_msg(self._connection.send_and_get_reply(
                msg, timeout=timeout
            ))

        return inner


def unwrap_read(b):
    """Raise ConnectionResetError from an empty read.

    Sometimes the socket raises an error itself, sometimes it gives no data.
    I haven't worked out when it behaves each way.
    """
    if not b:
        raise ConnectionResetError(ECONNRESET, os.strerror(ECONNRESET))
    return b


def prep_socket(addr, enable_fds=False, timeout=2.0) -> socket.socket:
    """Create a socket and authenticate ready to send D-Bus messages"""
    sock = socket.socket(family=socket.AF_UNIX)

    # To impose the overall auth timeout, we'll update the timeout on the socket
    # before each send/receive. This is ugly, but we can't use the socket for
    # anything else until this has succeeded, so this should be safe.
    deadline = timeout_to_deadline(timeout)
    def with_sock_deadline(meth, *args):
        sock.settimeout(deadline_to_timeout(deadline))
        return meth(*args)

    try:
        with_sock_deadline(sock.connect, addr)
        authr = Authenticator(enable_fds=enable_fds)
        for req_data in authr:
            with_sock_deadline(sock.sendall, req_data)
            authr.feed(unwrap_read(with_sock_deadline(sock.recv, 1024)))
        with_sock_deadline(sock.sendall, BEGIN)
    except socket.timeout as e:
        sock.close()
        raise TimeoutError(f"Did not authenticate in {timeout} seconds") from e
    except:
        sock.close()
        raise

    sock.settimeout(None)  # Put the socket back in blocking mode
    return sock


def open_dbus_connection(
        bus='SESSION', enable_fds=False, auth_timeout=1.,
) -> DBusConnection:
    """Connect to a D-Bus message bus

    Pass ``enable_fds=True`` to allow sending & receiving file descriptors.
    An error will be raised if the bus does not allow this. For simplicity,
    it's advisable to leave this disabled unless you need it.

    D-Bus has an authentication step before sending or receiving messages.
    This takes < 1 ms in normal operation, but there is a timeout so that client
    code won't get stuck if the server doesn't reply. *auth_timeout* configures
    this timeout in seconds.
    """
    bus_addr = get_bus(bus)
    sock = prep_socket(bus_addr, enable_fds, timeout=auth_timeout)

    conn = DBusConnection(sock, enable_fds)
    return conn


if __name__ == '__main__':
    conn = open_dbus_connection()
    print("Unique name:", conn.unique_name)



import trio

from jeepney import DBusAddress, new_method_call
from jeepney.bus_messages import message_bus, MatchRule
from jeepney.io.trio import open_dbus_router, Proxy

portal = DBusAddress(
    object_path='/org/freedesktop/portal/desktop',
    bus_name='org.freedesktop.portal.Desktop',
)
filechooser = portal.with_interface('org.freedesktop.portal.FileChooser')
token = 'file_choice_1'

async def choose_file():
    async with open_dbus_router() as router:
        sender_name = router.unique_name[1:].replace('.', '_')
        handle = f"/org/freedesktop/portal/desktop/request/{sender_name}/{token}"

        response_rule = MatchRule(
            type='signal', interface='org.freedesktop.portal.Request', path=handle
        )
        await Proxy(message_bus, router).AddMatch(response_rule)

        async with router.filter(response_rule) as responses:
            # https://flatpak.github.io/xdg-desktop-portal/portal-docs.html#gdbus-method-org-freedesktop-portal-FileChooser.OpenFile
            req = new_method_call(filechooser, 'OpenFile', 'ssa{sv}', (
                # Parent window, title, options
                '', 'Pick a file', {'handle_token': ('s', token)}
            ))
            await router.send_and_get_reply(req)

            response_signal_msg = await responses.receive()

        response, results = response_signal_msg.body
        if response == 0:
            # print(results)
            print("Chose file:", results['uris'][1][0])

if __name__ == '__main__':
    trio.run(choose_file)



"""
Example of subscribing to a D-Bus signal using blocking I/O.
This subscribes to the signal for a desktop notification being closed.

To try it, start this script, then trigger a desktop notification, and close it
somehow to trigger the signal. Use Ctrl-C to stop the script.

This example relies on the ``org.freedesktop.Notifications.NotificationClosed``
signal; some desktops may not support it. See the notification spec for more
details:
https://people.gnome.org/~mccann/docs/notification-spec/notification-spec-latest.html

Match rules are defined in the D-Bus specification:
https://dbus.freedesktop.org/doc/dbus-specification.html#message-bus-routing-match-rules
"""
from collections import deque

from jeepney.bus_messages import MatchRule, message_bus
from jeepney.io.blocking import open_dbus_connection, Proxy
from jeepney.wrappers import DBusAddress

noti = DBusAddress('/org/freedesktop/Notifications',
                   bus_name='org.freedesktop.Notifications',
                   interface='org.freedesktop.Notifications')

connection = open_dbus_connection(bus="SESSION")

match_rule = MatchRule(
    type="signal",
    interface=noti.interface,
    member="NotificationClosed",
    path=noti.object_path,
)



# Tell the session bus to pass us matching signal messages:
bus_proxy = Proxy(message_bus, connection)
print("Match added?", bus_proxy.AddMatch(match_rule) == ())

print("Trigger a desktop notification (e.g. with notify-send) and then close it")

with connection.filter(match_rule) as queue:
    signal_msg = connection.recv_until_filtered(queue)

reasons = {1: 'expiry', 2: 'dismissal', 3: 'dbus', '4': 'undefined'}

nid, reason_no = signal_msg.body
reason = reasons.get(reason_no, 'unknown')
print('Notification {} closed by: {}'.format(nid, reason))

connection.close()



"""Demonstrate receiving a file descriptor

Start this, and then run one of the _send_fd.py scripts to send requests.
"""
from datetime import datetime
import signal
from threading import Thread

from jeepney import MessageType, HeaderFields, new_method_return, new_error
from jeepney.bus_messages import message_bus
from jeepney.io.threading import (
    open_dbus_connection, DBusRouter, Proxy, ReceiveStopped,
)

SERVER_NAME = "io.gitlab.takluyver.jeepney.examples.FDWriter"

def serve(conn, i):
    while True:
        try:
            msg = conn.receive()
        except ReceiveStopped:
            return

        if msg.header.message_type != MessageType.method_call:
            print("Received non-method-call message:", msg)

        method = msg.header.fields[HeaderFields.member]
        print(f"Thread {i}: Message {msg.header.serial} calls {method}")

        if method == 'write_data':
            # body contains a FileDescriptor object, which we can convert to a file:
            fd, = msg.body
            with fd.to_file('w') as f:
                f.write(f'Timestamp: {datetime.now()}, server thread {i}')
            # Leaving the with block will close the fd in this process

            rep = new_method_return(msg, '')  # Empty reply to say we're done
        else:
            rep = new_error(msg, SERVER_NAME + '.Error.NoMethod')

        conn.send(rep)


with open_dbus_connection(enable_fds=True) as conn:
    # Request an additional name on the message bus
    with DBusRouter(conn) as router:
        bus_proxy = Proxy(message_bus, router, timeout=10)
        if bus_proxy.RequestName(SERVER_NAME) == (1,):
            # 1 == DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER
            print("Got name", SERVER_NAME)

    threads = [Thread(target=serve, args=(conn, i)) for i in range(4)]
    for t in threads:
        t.start()

    try:
        signal.pause()  # Wait for Ctrl-C
    except KeyboardInterrupt:
        pass

    conn.interrupt()
    for t in threads:
        t.join()




"""Make overlapping requests from several threads.

Use this alongside one of the _server.py examples
"""
import random
from threading import Thread

from jeepney import DBusAddress, new_method_call
from jeepney.io.threading import open_dbus_router

server = DBusAddress(
    "/io/gitlab/takluyver/jeepney/examples/Server",
    bus_name="io.gitlab.takluyver.jeepney.examples.Server",
)

def requests(router, i):
    for _ in range(5):
        n = random.randint(0, 5)
        # Construct a new D-Bus message. new_method_call takes the address, the
        # method name, the signature string, and a tuple of arguments.
        msg = new_method_call(server, 'slow_double', 'i', (n,))
        print(f"Thread {i} request:", n)

        # Send the message and wait for the reply
        reply = router.send_and_get_reply(msg)
        print(f'Thread {i} result:', reply.body[0])

with open_dbus_router() as router:
    threads = [Thread(target=requests, args=(router, i)) for i in range(6)]
    for t in threads:
        t.start()

    for t in threads:
        t.join()



"""Do not copy this example!

This is a deliberately broken example to test how D-Bus behaves when messages
are sent faster than they are handled. See other files in this folder for
useful examples of sending D-Bus messages nicely.
"""
import threading
import time

from jeepney import DBusAddress, new_method_call
from jeepney.io.blocking import open_dbus_connection

server = DBusAddress(
    "/io/gitlab/takluyver/jeepney/examples/Server",
    bus_name="io.gitlab.takluyver.jeepney.examples.Server",
)

connection = open_dbus_connection(bus='SESSION')
sending = True

def sender():
    i = 0
    while sending:
        i += 1
        n = i #random.randint(0, 5)
        # Construct a new D-Bus message. new_method_call takes the address, the
        # method name, the signature string, and a tuple of arguments.
        msg = new_method_call(server, 'slow_double', 'is', (n, 'a' * 1024 * 300))
        print(f"Request {i}: {n}")

        connection.send(msg)
        time.sleep(0.1)

send_thread = threading.Thread(target=sender)
send_thread.start()

def receiver():
    while True:
        msg = connection.receive()
        print('Result:', msg.body[0])

# recv_thread = threading.Thread(target=receiver)
# recv_thread.start()

try:
    receiver()
except KeyboardInterrupt:
    pass

sending = False
send_thread.join()
connection.close()



from warnings import warn

from .low_level import MessageType, HeaderFields
from .wrappers import DBusErrorResponse

class Router:
    """Routing for messages coming back to a client application.
    
    :param handle_factory: Constructor for an object like asyncio.Future,
        with methods *set_result* and *set_exception*. Outgoing method call
        messages will get a handle associated with them.
    :param on_unhandled: Callback for messages not otherwise dispatched.
    """
    def __init__(self, handle_factory, on_unhandled=None):
        self.handle_factory = handle_factory
        self._on_unhandled = on_unhandled
        self.outgoing_serial = 0
        self.awaiting_reply = {}
        self.signal_callbacks = {}

    @property
    def on_unhandled(self):
        return self._on_unhandled

    @on_unhandled.setter
    def on_unhandled(self, value):
        warn("Setting on_unhandled is deprecated. Please use the filter() "
             "method or simple receive() calls instead.", stacklevel=2)
        self._on_unhandled = value

    def outgoing(self, msg):
        """Set the serial number in the message & make a handle if a method call
        """
        self.outgoing_serial += 1
        msg.header.serial = self.outgoing_serial

        if msg.header.message_type is MessageType.method_call:
            self.awaiting_reply[msg.header.serial] = handle = self.handle_factory()
            return handle

    def subscribe_signal(self, callback, path, interface, member):
        """Add a callback for a signal.
        """
        warn("The subscribe_signal() method is deprecated. "
             "Please use the filter() API instead.", stacklevel=2)
        self.signal_callbacks[(path, interface, member)] = callback

    def incoming(self, msg):
        """Route an incoming message.
        """
        hdr = msg.header

        # Signals:
        if hdr.message_type is MessageType.signal:
            key = (hdr.fields.get(HeaderFields.path, None),
                   hdr.fields.get(HeaderFields.interface, None),
                   hdr.fields.get(HeaderFields.member, None)
                  )
            cb = self.signal_callbacks.get(key, None)
            if cb is not None:
                cb(msg.body)
                return

        # Method returns & errors
        reply_serial = hdr.fields.get(HeaderFields.reply_serial, -1)
        reply_handle = self.awaiting_reply.pop(reply_serial, None)
        if reply_handle is not None:
            if hdr.message_type is MessageType.method_return:
                reply_handle.set_result(msg.body)
                return
            elif hdr.message_type is MessageType.error:
                reply_handle.set_exception(DBusErrorResponse(msg))
                return

        if self.on_unhandled:
            self.on_unhandled(msg)



"""Simple client for the server in blocking_server.py
"""
import random
from time import sleep

from jeepney import DBusAddress, new_method_call
from jeepney.io.blocking import open_dbus_connection

server = DBusAddress(
    "/io/gitlab/takluyver/jeepney/examples/Server",
    bus_name="io.gitlab.takluyver.jeepney.examples.Server",
)

with open_dbus_connection() as connection:
    for i in range(10):
        n = random.randint(0, 5)
        # Construct a new D-Bus message. new_method_call takes the address, the
        # method name, the signature string, and a tuple of arguments.
        msg = new_method_call(server, 'double', 'i', (n,))
        print("Request:", n)

        # Send the message and wait for the reply
        reply = connection.send_and_get_reply(msg)
        print('Result:', reply.body[0])
        sleep(1)



import pytest
from jeepney.low_level import *

HELLO_METHOD_CALL = (
    b'l\x01\x00\x01\x00\x00\x00\x00\x01\x00\x00\x00m\x00\x00\x00\x01\x01o\x00\x15'
    b'\x00\x00\x00/org/freedesktop/DBus\x00\x00\x00\x02\x01s\x00\x14\x00\x00\x00'
    b'org.freedesktop.DBus\x00\x00\x00\x00\x03\x01s\x00\x05\x00\x00\x00Hello\x00'
    b'\x00\x00\x06\x01s\x00\x14\x00\x00\x00org.freedesktop.DBus\x00\x00\x00\x00')


def test_parser_simple():
    msg = Parser().feed(HELLO_METHOD_CALL)[0]
    assert msg.header.fields[HeaderFields.member] == 'Hello'

def chunks(src, size):
    pos = 0
    while pos < len(src):
        end = pos + size
        yield src[pos:end]
        pos = end

def test_parser_chunks():
    p = Parser()
    chunked = list(chunks(HELLO_METHOD_CALL, 16))
    for c in chunked[:-1]:
        assert p.feed(c) == []
    msg = p.feed(chunked[-1])[0]
    assert msg.header.fields[HeaderFields.member] == 'Hello'

def test_multiple():
    msgs = Parser().feed(HELLO_METHOD_CALL * 6)
    assert len(msgs) == 6
    for msg in msgs:
        assert msg.header.fields[HeaderFields.member] == 'Hello'

def test_roundtrip():
    msg = Parser().feed(HELLO_METHOD_CALL)[0]
    assert msg.serialise() == HELLO_METHOD_CALL

def test_serialise_dict():
    data = {
        'a': 'b',
        'de': 'f',
    }
    string_type = simple_types['s']
    sig = Array(DictEntry([string_type, string_type]))
    print(sig.serialise(data, 0, Endianness.little))
    assert sig.serialise(data, 0, Endianness.little) == (
        b'\x1e\0\0\0' +  # Length
        b'\0\0\0\0' +  # Padding
        b'\x01\0\0\0a\0\0\0' +
        b'\x01\0\0\0b\0\0\0' +
        b'\x02\0\0\0de\0\0' +
        b'\x01\0\0\0f\0'
    )

def test_parse_signature():
    sig = parse_signature(list('(a{sv}(oayays)b)'))
    print(sig)
    assert sig == Struct([
        Array(DictEntry([simple_types['s'], Variant()])),
        Struct([
            simple_types['o'],
            Array(simple_types['y']),
            Array(simple_types['y']),
            simple_types['s']
        ]),
        simple_types['b'],
    ])

class fake_list(list):
    def __init__(self, n):
        super().__init__()
        self._n = n

    def __len__(self):
        return self._n

    def __iter__(self):
        return iter(range(self._n))

def test_array_limit():
    # The spec limits arrays to 64 MiB
    a = Array(FixedType(8, 'Q'))  # 'at' - array of uint64
    a.serialise(fake_list(100), 0, Endianness.little)
    with pytest.raises(SizeLimitError):
        a.serialise(fake_list(2**23 + 1), 0, Endianness.little)



"""Pick a colour on your screen, show its details in the terminal

This probably only works on Gnome desktops, as it uses a Gnome D-Bus interface.
This is just an example, so broad support is less important than simplicity.
"""
from jeepney.io.blocking import open_dbus_connection, Proxy
from jeepney.wrappers import MessageGenerator, new_method_call

# This class was auto-generated with:
# python3 -m jeepney.bindgen --name org.gnome.Shell.Screenshot --path /org/gnome/Shell/Screenshot
class Screenshot(MessageGenerator):
    interface = 'org.gnome.Shell.Screenshot'

    def __init__(self, object_path='/org/gnome/Shell/Screenshot',
                 bus_name='org.gnome.Shell.Screenshot'):
        super().__init__(object_path=object_path, bus_name=bus_name)

    def Screenshot(self, include_cursor, flash, filename):
        return new_method_call(self, 'Screenshot', 'bbs',
                               (include_cursor, flash, filename))

    def ScreenshotWindow(self, include_frame, include_cursor, flash, filename):
        return new_method_call(self, 'ScreenshotWindow', 'bbbs',
                               (include_frame, include_cursor, flash, filename))

    def ScreenshotArea(self, x, y, width, height, flash, filename):
        return new_method_call(self, 'ScreenshotArea', 'iiiibs',
                               (x, y, width, height, flash, filename))

    def PickColor(self):
        return new_method_call(self, 'PickColor')

    def FlashArea(self, x, y, width, height):
        return new_method_call(self, 'FlashArea', 'iiii',
                               (x, y, width, height))

    def SelectArea(self):
        return new_method_call(self, 'SelectArea')


with open_dbus_connection() as conn:
    screenshot = Proxy(Screenshot(), conn)
    res_dict, = screenshot.PickColor()

# We get RGB as floats (0-1), and convert to integers (0-255)
rf, gf, bf = res_dict['color'][1]
ri, gi, bi = round(rf * 255), round(gf * 255), round(bf * 255)

print(f"RGB (0-1)  : ({rf:.5f}, {gf:.5f}, {bf:.5f})")
print(f"RGB (0-255): {ri, gi, bi}")
print(f"Hex code   : #{ri:02X}{gi:02X}{bi:02X}")
print(f"Preview    : \x1b[48;2;{ri};{gi};{bi}m            \x1b[0m")
# The line above assumes our terminal supports 24-bit colour:
# https://en.wikipedia.org/wiki/ANSI_escape_code#24-bit



from jeepney import DBusAddress, new_method_call
from jeepney.io.blocking import open_dbus_connection

notifications = DBusAddress('/org/freedesktop/Notifications',
                            bus_name='org.freedesktop.Notifications',
                            interface='org.freedesktop.Notifications')

connection = open_dbus_connection(bus='SESSION')

# Construct a new D-Bus message. new_method_call takes the address, the
# method name, the signature string, and a tuple of arguments.
msg = new_method_call(notifications, 'Notify', 'susssasa{sv}i',
                      ('jeepney_test',  # App name
                       0,   # Not replacing any previous notification
                       '',  # Icon
                       'Hello, world!',  # Summary
                       'This is an example notification from Jeepney',
                       [], {},  # Actions, hints
                       -1,      # expire_timeout (-1 = default)
                      ))

# Send the message and wait for the reply
reply = connection.send_and_get_reply(msg)
print('Notification ID:', reply.body[0])

connection.close()



import asyncio

from jeepney.io.asyncio import open_dbus_connection

async def hello():
    conn = await open_dbus_connection(bus='SESSION')
    print('My ID is:', conn.unique_name)
    
asyncio.run(hello())



"""Send a desktop notification

See also aio_notify.py, which does the same with the higher-level Proxy API.
"""
import asyncio

from jeepney import DBusAddress, new_method_call
from jeepney.io.asyncio import open_dbus_router

notifications = DBusAddress('/org/freedesktop/Notifications',
                            bus_name='org.freedesktop.Notifications',
                            interface='org.freedesktop.Notifications')

async def send_notification():
    msg = new_method_call(notifications, 'Notify', 'susssasa{sv}i',
                            ('jeepney_test',  # App name
                             0,     # Not replacing any previous notification
                             '',    # Icon
                             'Hello, world!',  # Summary
                             'This is an example notification from Jeepney',
                             [], {},  # Actions, hints
                             -1,      # expire_timeout (-1 = default)
                            ))
    # Send the message and await the reply
    async with open_dbus_router() as router:
        reply = await router.send_and_get_reply(msg)
        print('Notification ID:', reply.body[0])

loop = asyncio.get_event_loop()
loop.run_until_complete(send_notification())



from jeepney import DBusAddress, new_signal, new_method_call
from jeepney.bus_messages import MatchRule, message_bus

portal = DBusAddress(
    object_path='/org/freedesktop/portal/desktop',
    bus_name='org.freedesktop.portal.Desktop',
)
portal_req_iface = portal.with_interface('org.freedesktop.portal.Request')


def test_match_rule_simple():
    rule = MatchRule(
        type='signal', interface='org.freedesktop.portal.Request',
    )
    assert rule.matches(new_signal(portal_req_iface, 'Response'))

    # Wrong message type
    assert not rule.matches(new_method_call(portal_req_iface, 'Boo'))

    # Wrong interface
    assert not rule.matches(new_signal(
        portal.with_interface('org.freedesktop.portal.FileChooser'), 'Response'
    ))


def test_match_rule_path_namespace():
    assert MatchRule(path_namespace='/org/freedesktop/portal').matches(
        new_signal(portal_req_iface, 'Response')
    )

    # Prefix but not a parent in the path hierarchy
    assert not MatchRule(path_namespace='/org/freedesktop/por').matches(
        new_signal(portal_req_iface, 'Response')
    )


def test_match_rule_arg():
    rule = MatchRule(type='method_call')
    rule.add_arg_condition(0, 'foo')

    assert rule.matches(new_method_call(
        portal_req_iface, 'Boo', signature='s', body=('foo',)
    ))

    assert not rule.matches(new_method_call(
        portal_req_iface, 'Boo', signature='s', body=('foobar',)
    ))

    # No such argument
    assert not rule.matches(new_method_call(portal_req_iface, 'Boo'))


def test_match_rule_arg_path():
    rule = MatchRule(type='method_call')
    rule.add_arg_condition(0, '/aa/bb/', kind='path')

    # Exact match
    assert rule.matches(new_method_call(
        portal_req_iface, 'Boo', signature='s', body=('/aa/bb/',)
    ))

    # Match a prefix
    assert rule.matches(new_method_call(
        portal_req_iface, 'Boo', signature='s', body=('/aa/bb/cc',)
    ))

    # Argument is a prefix, ending with /
    assert rule.matches(new_method_call(
        portal_req_iface, 'Boo', signature='s', body=('/aa/',)
    ))

    # Argument is a prefix, but NOT ending with /
    assert not rule.matches(new_method_call(
        portal_req_iface, 'Boo', signature='s', body=('/aa',)
    ))

    assert not rule.matches(new_method_call(
        portal_req_iface, 'Boo', signature='s', body=('/aa/bb',)
    ))

    # Not a string
    assert not rule.matches(new_method_call(
        portal_req_iface, 'Boo', signature='u', body=(12,)
    ))


def test_match_rule_arg_namespace():
    rule = MatchRule(member='NameOwnerChanged')
    rule.add_arg_condition(0, 'com.example.backend1', kind='namespace')

    # Exact match
    assert rule.matches(new_signal(
        message_bus, 'NameOwnerChanged', 's', ('com.example.backend1',)
    ))

    # Parent of the name
    assert rule.matches(new_signal(
        message_bus, 'NameOwnerChanged', 's', ('com.example.backend1.foo.bar',)
    ))

    # Prefix but not a parent in the namespace
    assert not rule.matches(new_signal(
        message_bus, 'NameOwnerChanged', 's', ('com.example.backend12',)
    ))

    # Not a string
    assert not rule.matches(new_signal(
        message_bus, 'NameOwnerChanged', 'u', (1,)
    ))



"""Example accessing the SecretStorage DBus interface with blocking APIs

https://freedesktop.org/wiki/Specifications/secret-storage-spec/secrets-api-0.1.html#ref-dbus-api
"""

from jeepney import new_method_call, DBusAddress, Properties
from jeepney.io.blocking import open_dbus_connection

secrets = DBusAddress('/org/freedesktop/secrets',
                           bus_name= 'org.freedesktop.secrets',
                           interface='org.freedesktop.Secret.Service')

login_keyring = DBusAddress('/org/freedesktop/secrets/collection/login',
                           bus_name= 'org.freedesktop.secrets',
                           interface='org.freedesktop.Secret.Collection')

list_items = new_method_call(login_keyring, 'SearchItems', 'a{ss}',
                             ([],))


conn = open_dbus_connection(bus='SESSION')

resp = conn.send_and_get_reply(Properties(secrets).get('Collections'))
print('Collections:', resp.body[0][1])

print('\nItems in login collection:')
all_items = conn.send_and_get_reply(list_items).body[0]
for obj_path in all_items:
    item = DBusAddress(obj_path, 'org.freedesktop.secrets',
                      interface='org.freedesktop.Secret.Item')
    props_resp = conn.send_and_get_reply(Properties(item).get_all())
    props = dict(props_resp.body[0])
    state = '(locked)' if props['Locked'][1] else ''
    print(props['Label'][1], state)

conn.close()



import sys

sys.excepthook =  sys.__excepthook__

import trio

from jeepney import MessageGenerator, new_method_call
from jeepney.io.trio import open_dbus_router, Proxy

# ---- Message generator, created by jeepney.bindgen ----
class Notifications(MessageGenerator):
    interface = 'org.freedesktop.Notifications'

    def __init__(self, object_path='/org/freedesktop/Notifications',
                 bus_name='org.freedesktop.Notifications'):
        super().__init__(object_path=object_path, bus_name=bus_name)

    def Notify(self, arg_0, arg_1, arg_2, arg_3, arg_4, arg_5, arg_6, arg_7):
        return new_method_call(self, 'Notify', 'susssasa{sv}i',
                               (arg_0, arg_1, arg_2, arg_3, arg_4, arg_5, arg_6, arg_7))

    def CloseNotification(self, arg_0):
        return new_method_call(self, 'CloseNotification', 'u',
                               (arg_0,))

    def GetCapabilities(self):
        return new_method_call(self, 'GetCapabilities')

    def GetServerInformation(self):
        return new_method_call(self, 'GetServerInformation')
# ---- End auto generated code ----


async def send_notification():
    async with open_dbus_router(bus='SESSION') as req:
        proxy = Proxy(Notifications(), req)

        resp = await proxy.Notify('jeepney_test',  # App name
                              0,      # Not replacing any previous notification
                              '',     # Icon
                              'Hello, world!',  # Summary
                              'This is an example notification from Jeepney & Trio',
                              [], {},  # Actions, hints
                              -1,      # expire_timeout (-1 = default)
                             )
        print('Notification ID:', resp[0])

if __name__ == '__main__':
    trio.run(send_notification)



"""Demonstrate sending a file descriptor

Start one of the _recv_fd.py scripts first, then run this.

Make a writable temporary file, send its file descriptor. The other process
writes into it before replying. This process can read what was written.
This works even though the temp file may never have a name on the filesystem
for the other process to open it.
"""
from tempfile import TemporaryFile

from jeepney import DBusAddress, new_method_call
from jeepney.io.blocking import open_dbus_connection

server = DBusAddress(
    "/io/gitlab/takluyver/jeepney/examples/FDWriter",
    bus_name="io.gitlab.takluyver.jeepney.examples.FDWriter",
)

with open_dbus_connection(enable_fds=True) as connection:
    with TemporaryFile() as tf:
        # TemporaryFile() has a .fileno() method, so it can be passed as the
        # data for a file descriptor (h for handle).
        msg = new_method_call(server, 'write_data', 'h', (tf,))
        print("Sending:", tf)

        reply = connection.send_and_get_reply(msg)

        # Retrieve what the other process has written
        tf.seek(0)
        print("File contents:", tf.read())



"""Demonstrate receiving a file descriptor

Start this, and then run one of the _send_fd.py scripts to send requests.
"""
from datetime import datetime

from jeepney import MessageType, HeaderFields, new_method_return, new_error
from jeepney.bus_messages import message_bus
from jeepney.io.blocking import open_dbus_connection

SERVER_NAME = "io.gitlab.takluyver.jeepney.examples.FDWriter"

with open_dbus_connection(enable_fds=True) as connection:
    print("My unique name is:", connection.unique_name)

    # Request an additional name on the message bus
    rep = connection.send_and_get_reply(message_bus.RequestName(SERVER_NAME))
    if rep.body[0] == 1:  # DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER
        print("Got name", SERVER_NAME)

    while True:
        msg = connection.receive()
        if msg.header.message_type != MessageType.method_call:
            print("Received non-method-call message:", msg)

        method = msg.header.fields[HeaderFields.member]
        print(f"Message {msg.header.serial} calls {method}")

        if method == 'write_data':
            # body contains a FileDescriptor object, which we can convert to a file:
            fd, = msg.body
            with fd.to_file('w') as f:
                f.write(f'Timestamp: {datetime.now()}')
            # Leaving the with block will close the fd in this process

            rep = new_method_return(msg, '')  # Empty reply to say we're done
        else:
            rep = new_error(msg, SERVER_NAME + '.Error.NoMethod')

        connection.send_message(rep)




"""
This example simulates two clients interacting with the message bus, more or
less independently.

Client 1
    Any app on the bus (here, called *service*). It asks the bus for sole
    custody of its preferred name, a `well-known bus name`_ that it wants
    others to recognize.

Client 2
    An interested party (here, called *watcher*). It asks the bus to emit a
    signal (send it an update) whenever *service*'s well-known bus name changes
    hands.  It uses Jeepney's ``router.filter()`` to get a message whenever this
    happens.

.. _well-known bus name: https://dbus.freedesktop.org/doc
   /dbus-specification.html#message-protocol-names
"""

import asyncio

from jeepney.io.asyncio import open_dbus_router, Proxy
from jeepney.bus_messages import message_bus, MatchRule

well_known_bus_name = "io.readthedocs.jeepney.aio_subscribe_example"

async def watcher(ready: asyncio.Event):
    # Create a "signal-selection" match rule
    match_rule = MatchRule(
        type="signal",
        sender=message_bus.bus_name,
        interface=message_bus.interface,
        member="NameOwnerChanged",
        path=message_bus.object_path,
    )

    # Condition: arg number 0 must match the bus name (try changing either)
    match_rule.add_arg_condition(0, well_known_bus_name)

    async with open_dbus_router() as router:
        await Proxy(message_bus, router).AddMatch(match_rule)

        with router.filter(match_rule) as q:
            print("[watcher] subscribed to NameOwnerChanged signal")
            ready.set()

            msg = await q.get()
            print("[watcher] match hit:", msg.body)

async def main():
    watcher_ready = asyncio.Event()
    watcher_task = asyncio.create_task(watcher(watcher_ready))
    await watcher_ready.wait()

    async with open_dbus_router() as router:
        print("[service] calling 'RequestName'")
        resp = await Proxy(message_bus, router).RequestName(well_known_bus_name, 4)
        print("[service] reply:", (None, "primary owner", "in queue",
                                   "exists", "already owned")[resp[0]])

    await watcher_task

if __name__ == "__main__":
    asyncio.run(main())



"""Demonstrate receiving a file descriptor

Start this, and then run one of the _send_fd.py scripts to send requests.
"""
from datetime import datetime

import trio

from jeepney import MessageType, HeaderFields, new_method_return, new_error
from jeepney.bus_messages import message_bus
from jeepney.io.trio import (
    open_dbus_connection, Proxy,
)

SERVER_NAME = "io.gitlab.takluyver.jeepney.examples.FDWriter"

async def serve(conn, i):
    while True:
        msg = await conn.receive()

        if msg.header.message_type != MessageType.method_call:
            print("Received non-method-call message:", msg)
            continue

        method = msg.header.fields[HeaderFields.member]
        print(f"Task {i}: Message {msg.header.serial} calls {method}")

        if method == 'write_data':
            # body contains a FileDescriptor object, which we can convert to a file:
            fd, = msg.body
            with fd.to_file('w') as f:
                f.write(f'Timestamp: {datetime.now()}, server task {i}')
            # Leaving the with block will close the fd in this process

            rep = new_method_return(msg, '')  # Empty reply to say we're done
        else:
            rep = new_error(msg, SERVER_NAME + '.Error.NoMethod')

        await conn.send(rep)

async def main():
    conn = await open_dbus_connection(enable_fds=True)
    # Request an additional name on the message bus
    async with conn.router() as router:
        bus_proxy = Proxy(message_bus, router)
        with trio.fail_after(2):
            reply, = await bus_proxy.RequestName(SERVER_NAME)
            if reply == 1:
                # 1 == DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER
                print("Got name", SERVER_NAME)

    async with trio.open_nursery() as nursery:
        for i in range(4):
            nursery.start_soon(serve, conn, i)

trio.run(main)




from jeepney import DBusAddress, new_method_call
from jeepney.bus_messages import message_bus, MatchRule
from jeepney.io.blocking import open_dbus_connection, Proxy

portal = DBusAddress(
    object_path='/org/freedesktop/portal/desktop',
    bus_name='org.freedesktop.portal.Desktop',
)
filechooser = portal.with_interface('org.freedesktop.portal.FileChooser')

conn = open_dbus_connection()

token = 'file_choice_1'
sender_name = conn.unique_name[1:].replace('.', '_')
handle = f"/org/freedesktop/portal/desktop/request/{sender_name}/{token}"

response_rule = MatchRule(
    type='signal', interface='org.freedesktop.portal.Request', path=handle
)
Proxy(message_bus, conn).AddMatch(response_rule)

with conn.filter(response_rule) as responses:
    # https://flatpak.github.io/xdg-desktop-portal/portal-docs.html#gdbus-method-org-freedesktop-portal-FileChooser.OpenFile
    req = new_method_call(filechooser, 'OpenFile', 'ssa{sv}', (
        # Parent window, title, options
        '', 'Pick a file', {'handle_token': ('s', token)}
    ))
    conn.send_and_get_reply(req)

    # responses is a deque - process messages until one is filtered into it
    response_msg = conn.recv_until_filtered(responses)

response, results = response_msg.body
if response == 0:
    # print(results)
    print("Chose file:", results['uris'][1][0])

conn.close()



"""A simple server built on Jeepney's blocking integration.

Start this, and then run blocking_server_client.py to send requests to it.
"""
import signal
from threading import Thread
import time

from jeepney import MessageType, HeaderFields, new_method_return, new_error
from jeepney.bus_messages import message_bus
from jeepney.io.threading import (
    open_dbus_connection, DBusRouter, Proxy, ReceiveStopped,
)

SERVER_NAME = "io.gitlab.takluyver.jeepney.examples.Server"

def slow_double(n):
    # Simulate something slow which releases the GIL so other threads can run
    time.sleep(5)
    return n * 2

def double(n):
    return n * 2

def divide(n, d):
    return n / d


def serve(conn, i):
    while True:
        try:
            msg = conn.receive()
        except ReceiveStopped:
            return

        if msg.header.message_type != MessageType.method_call:
            print("Received non-method-call message:", msg)

        method = msg.header.fields[HeaderFields.member]
        print(f"Thread {i}: Message {msg.header.serial} calls {method}")

        # Dispatch to different methods
        if method == 'slow_double':
            res = slow_double(msg.body[0])
            print(f"Delayed reply to {msg.header.serial}: {res}")
            rep = new_method_return(msg, 'i', (res,))

        elif method == 'double':
            res = double(msg.body[0])
            rep = new_method_return(msg, 'i', (res,))

        elif method == 'divide':
            try:
                res = divide(*msg.body)
            except ZeroDivisionError:
                rep = new_error(msg, SERVER_NAME + ".Error.DivideByZero")
            else:
                rep = new_method_return(msg, 'd', (res,))

        else:
            rep = new_error(msg, SERVER_NAME + '.Error.NoMethod')

        conn.send(rep)


with open_dbus_connection() as conn:
    print("My unique name is:", conn.unique_name)

    # Request an additional name on the message bus
    with DBusRouter(conn) as router:
        bus_proxy = Proxy(message_bus, router, timeout=10)
        if bus_proxy.RequestName(SERVER_NAME) == (1,):
            # 1 == DBUS_REQUEST_NAME_REPLY_PRIMARY_OWNER
            print("Got name", SERVER_NAME)

    threads = [Thread(target=serve, args=(conn, i)) for i in range(4)]
    for t in threads:
        t.start()

    try:
        signal.pause()  # Wait for Ctrl-C
    except KeyboardInterrupt:
        pass

    conn.interrupt()
    for t in threads:
        t.join()




"""Demonstrate sending a file descriptor

Start one of the _recv_fd.py scripts first, then run this.

Make a writable temporary file, send its file descriptor. The other process
writes into it before replying. This process can read what was written.
This works even though the temp file may never have a name on the filesystem
for the other process to open it.
"""
from tempfile import TemporaryFile
from threading import Thread
import time

from jeepney import DBusAddress, new_method_call
from jeepney.io.threading import open_dbus_router

server = DBusAddress(
    "/io/gitlab/takluyver/jeepney/examples/FDWriter",
    bus_name="io.gitlab.takluyver.jeepney.examples.FDWriter",
)

def requests(router, i):
    for _ in range(4):
        with TemporaryFile() as tf:
            # Construct a new D-Bus message. new_method_call takes the address, the
            # method name, the signature string, and a tuple of arguments.
            msg = new_method_call(server, 'write_data', 'h', (tf,))
            print(f"Client thread {i} sending", tf)

            # Send the message and wait for the reply
            router.send_and_get_reply(msg)

            # Retrieve what the other process has written
            tf.seek(0)
            print(f"Client thread {i} reads:", tf.read())

        time.sleep(0.5)

with open_dbus_router(enable_fds=True) as router:
    threads = [Thread(target=requests, args=(router, i)) for i in range(5)]
    for t in threads:
        t.start()

    for t in threads:
        t.join()



import array
import os
import socket
from warnings import warn


class NoFDError(RuntimeError):
    """Raised by :class:`FileDescriptor` methods if it was already closed/converted
    """
    pass


class FileDescriptor:
    """A file descriptor received in a D-Bus message

    This wrapper helps ensure that the file descriptor is closed exactly once.
    If you don't explicitly convert or close the FileDescriptor object, it will
    close its file descriptor when it goes out of scope, and emit a
    ResourceWarning.
    """
    __slots__ = ('_fd',)
    _CLOSED = -1
    _CONVERTED = -2

    def __init__(self, fd):
        self._fd = fd

    def __repr__(self):
        detail = self._fd
        if self._fd == self._CLOSED:
            detail = 'closed'
        elif self._fd == self._CONVERTED:
            detail = 'converted'
        return f"<FileDescriptor ({detail})>"

    def close(self):
        """Close the file descriptor

        This can safely be called multiple times, but will raise RuntimeError
        if called after converting it with one of the ``to_*`` methods.

        This object can also be used in a ``with`` block, to close it on
        leaving the block.
        """
        if self._fd == self._CLOSED:
            pass
        elif self._fd == self._CONVERTED:
            raise NoFDError("Can't close FileDescriptor after converting it")
        else:
            self._fd, fd = self._CLOSED, self._fd
            os.close(fd)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        if self._fd >= 0:
            warn(
                f'FileDescriptor ({self._fd}) was neither closed nor converted',
                ResourceWarning, stacklevel=2, source=self
            )
            self.close()

    def _check(self):
        if self._fd < 0:
            detail = 'closed' if self._fd == self._CLOSED else 'converted'
            raise NoFDError(f'FileDescriptor object was already {detail}')

    def fileno(self):
        """Get the integer file descriptor

        This does not change the state of the :class:`FileDescriptor` object,
        unlike the ``to_*`` methods.
        """
        self._check()
        return self._fd

    def to_raw_fd(self):
        """Convert to the low-level integer file descriptor::

            raw_fd = fd.to_raw_fd()
            os.write(raw_fd, b'xyz')
            os.close(raw_fd)

        The :class:`FileDescriptor` can't be used after calling this. The caller
        is responsible for closing the file descriptor.
        """
        self._check()
        self._fd, fd = self._CONVERTED, self._fd
        return fd

    def to_file(self, mode, buffering=-1, encoding=None, errors=None, newline=None):
        """Convert to a Python file object::

            with fd.to_file('w') as f:
                f.write('xyz')

        The arguments are the same as for the builtin :func:`open` function.

        The :class:`FileDescriptor` can't be used after calling this. Closing
        the file object will also close the file descriptor.
        """
        self._check()
        f = open(
            self._fd, mode, buffering=buffering,
            encoding=encoding, errors=errors, newline=newline
        )
        self._fd = self._CONVERTED
        return f

    def to_socket(self):
        """Convert to a socket object

        This returns a standard library :func:`socket.socket` object::

            with fd.to_socket() as sock:
                b = sock.sendall(b'xyz')

        The wrapper object can't be used after calling this. Closing the socket
        object will also close the file descriptor.
        """
        from socket import socket

        self._check()
        s = socket(fileno=self._fd)
        self._fd = self._CONVERTED
        return s

    @classmethod
    def from_ancdata(cls, ancdata) -> ['FileDescriptor']:
        """Make a list of FileDescriptor from received file descriptors

        ancdata is a list of ancillary data tuples as returned by socket.recvmsg()
        """
        fds = array.array("i")  # Array of ints
        for cmsg_level, cmsg_type, data in ancdata:
            if cmsg_level == socket.SOL_SOCKET and cmsg_type == socket.SCM_RIGHTS:
                # Append data, ignoring any truncated integers at the end.
                fds.frombytes(data[:len(data) - (len(data) % fds.itemsize)])
        return [cls(i) for i in fds]


_fds_buf_size_cache = None

def fds_buf_size():
    # If there may be file descriptors, we try to read 1 message at a time.
    # The reference implementation of D-Bus defaults to allowing 16 FDs per
    # message, and the Linux kernel currently allows 253 FDs per sendmsg()
    # call. So hopefully allowing 256 FDs per recvmsg() will always suffice.
    global _fds_buf_size_cache
    if _fds_buf_size_cache is None:
        maxfds = 256
        fd_size = array.array('i').itemsize
        _fds_buf_size_cache = socket.CMSG_SPACE(maxfds * fd_size)
    return _fds_buf_size_cache



"""Example accessing the SecretStorage DBus interface with asyncio APIs

https://freedesktop.org/wiki/Specifications/secret-storage-spec/secrets-api-0.1.html#ref-dbus-api
"""
import asyncio

from jeepney import new_method_call, DBusAddress, Properties
from jeepney.io.asyncio import open_dbus_router

secrets = DBusAddress('/org/freedesktop/secrets',
                      bus_name= 'org.freedesktop.secrets',
                      interface='org.freedesktop.Secret.Service')

login_keyring = DBusAddress('/org/freedesktop/secrets/collection/login',
                            bus_name= 'org.freedesktop.secrets',
                            interface='org.freedesktop.Secret.Collection')

search_msg = new_method_call(
    login_keyring, 'SearchItems', 'a{ss}', ({'user': 'tk2e15',},)
)

async def query_secrets():
    async with open_dbus_router() as router:
        get_collections_msg = Properties(secrets).get('Collections')
        resp = await router.send_and_get_reply(get_collections_msg)
        print('Collections:', resp.body[0][1])

        resp = await router.send_and_get_reply(search_msg)
        print('Search res:', resp.body)


asyncio.run(query_secrets())

